# RandomBots by David Bowland
# ./addons/eventscripts/randombots/randombots.py


# >>> To configure this addon please see randombots.cfg  (created when the addon is first loaded) <<<

"""
Randomly sets bot difficulty for each bot, making some bots better than others.
Requires only ES 2.0.0.250i+
"""

import cfglib
import cmdlib
import es
import gamethread
import random
from path import path


import psyco
psyco.full()

###

info = es.AddonInfo()
info.name      = 'RandomBots'
info.version   = '1'
info.basename  = 'randombots'
info.url       = 'http://addons.eventscripts.com/addons/view/' + info.basename
info.author    = 'SuperDave'

###

# Directory of the current game
game_path = path(es.ServerVar('eventscripts_gamedir'))
# Directory of this addon
addon_path = path(es.getAddonPath(info.basename))
# Path to existing bot profile database
db_path = game_path.joinpath('BotProfile.db')
# Path to RB bot profile database
db_path_rb = addon_path.joinpath('BotProfile.db')
# Path to existing db backup
backup_path = game_path.joinpath('BotProfile_rb.db')

###

config = cfglib.AddonCFG(addon_path.joinpath(info.basename + '.cfg'))

config.text(info.name + ' release %s options' % info.version)
config.text('./addons/eventscripts/%(basename)s/%(basename)s.cfg' % {'basename': info.basename})
config.text('')
config.text('Load this addon with: es_load ' + info.basename)
config.text('')
config.text('Bots can be made even more realistic with SloBot: http://addons.eventscripts.com/pages/addons/view/slobot')
config.text('\n')
config.text('The next bot difficulty level will be located in the server command "bot_difficulty"')
config.text('\n')

config.text('***** Difficulty weight options *****')
config.text('The following options determine the weight of each difficulty when selecting a random difficulty.')
config.text('Higher numbers for a difficulty mean a greater chance of selecting that difficulty.')
config.text('Being weighted options, these options are relative to each other.')
cvar_easy   = config.cvar('randombots_easy',   15, 'Weight of selecting an easy bot')
cvar_normal = config.cvar('randombots_normal', 40, 'Weight of selecting a normal bot')
cvar_hard   = config.cvar('randombots_hard',   30, 'Weight of selecting a hard bot')
cvar_expert = config.cvar('randombots_expert', 15, 'Weight of selecting an expert bot')
config.text('\n')

config.text('***** Bot database option *****')
config.text('While RandomBots will attempt to preserve existing data, if you choose to use the')
config.text('RB database it is advisable to save a copy of your existing database.')
cvar_use_database = config.cvar('randombots_use_database', 1, '0 = no change, 1 = use RandomBots database of bots (located in ./randombots/BotProfile.db)')
config.text('')
config.text('The RB database has been tweaked to bots a bit harder and more realistic.')
config.text('For example, easy level bots will no longer stare at an opponent for over a second before firing.')
config.text('\n')

config.text('***** Bot quota options *****')
cvar_quota_control = config.cvar('randombots_quota_control', 1, '0 = no change, 1 = RandomBots controls bot quota')
config.text('')
cvar_quota_default_min = config.cvar('randombots_quota_default_min', 8,  'Minimum number of bots on a map where no bot quota is specified')
cvar_quota_default_max = config.cvar('randombots_quota_default_max', 12, 'Maximum number of bots on a map where no bot quota is specified')
config.text('')
cvar_quota_delay_min = config.cvar('randombots_quota_delay_min', 75,  'Minimum number of seconds between changing the bot quota')
cvar_quota_delay_max = config.cvar('randombots_quota_delay_max', 210, 'Maximum number of seconds between changing the bot quota')
config.text('\n')

config.text('Use the following server command to specify bot quota values for specific maps:')
config.text('')
config.text('randombots_quota_add <map name> <min quota> <max quota>')
config.text('\n')

config.text('Place your randombots_quota_add commands below:')
config.command('randombots_quota_add')
config.text('\n')

config.text('Uncomment the following example options if you wish to use them:')
config.text('')
config.text('randombots_quota_add cs_assault 6 8')
config.text('randombots_quota_add de_dust 10 14')
config.text('\n')

config.text('Use the following command to remove custom bot quotas from a specific map:')
config.text('')
config.text('randombots_quota_remove <map name>')
config.text('')
config.text('Due to the bot quotas can be specified above, this command will not be used by most users.')
config.text('\n')

config.text('***** Server bot options *****')
config.text('You can set your other bot options below to keep bot options in a single location.')
config.text('Below is a list of notable bot cvars.')
config.text('')
config.text("If nonzero, bots may occasionally go 'rogue'. Rogue bots do not obey radio commands, nor pursue scenario goals.")
config.text('bot_allow_rogues 1')
config.text('')
config.text('If nonzero, bots will automatically leave to make room for human players.')
config.text('bot_auto_vacate 1')
config.text('')
config.text('If nonzero and there is a human on the team, the bots will not do the scenario tasks.')
config.text('bot_defer_to_human 0')
config.text('')
config.text('If nonzero, bots will not buy if their money falls below this amount.')
config.text('bot_eco_limit 750')
config.text('')
config.text('This string is prefixed to the name of all bots that join the game.')
config.text("<difficulty> will be replaced with the bot's difficulty.")
config.text("<weaponclass> will be replaced with the bot's desired weapon class.")
config.text("<skill> will be replaced with a 0-100 representation of the bot's skill.")
config.text('bot_prefix ""')
config.text('')
config.text("Allowed values: 'normal', 'fill', and 'match'.")
config.text("If 'fill', the server will adjust bots to keep N players in the game, where N is bot_quota.")
config.text("If 'match', the server will maintain a 1:N ratio of humans to bots, where N is bot_quota.")
config.text('bot_quota_mode "fill"')

config.write()


cvar_difficulty = es.ServerVar('bot_difficulty')
cvar_quota = es.ServerVar('bot_quota')
cvar_currentmap = es.ServerVar('eventscripts_currentmap')

###

def selectNextDifficulty():
   """ Sets the cvar bot_difficulty to a random difficulty """
   cvar_difficulty.set(random.choice(
    [0] * int(cvar_easy) +
    [1] * int(cvar_normal) +
    [2] * int(cvar_hard) +
    [3] * int(cvar_expert)))

###

def loadProfile():
   """ Loads the RandomBots profile database and picks a new bot difficulty """
   # If we don't need to use the RB profile database we're done here
   if not cvar_use_database:
      return

   # If the backup db exists we assume it's the original db.
   # If, for instance, the server crashes while RB is loaded,
   # we won't think the RB db is the original db.
   if db_path.isfile() and not backup_path.isfile():
      db_path.move(backup_path)

   # Now we copy the RB database where it will be used by Source
   db_path_rb.copyfile(db_path)


def unloadProfile():
   """ Reverts the bot profile database to its original state """
   # If the backup db exists we want to copy it to the function location
   # regardless of cvar_use_database. This is so we can revert anything we may
   # have accidentally done.
   if backup_path.isfile():
      # Move backup file back to the db location
      backup_path.move(db_path)

   # If there is not backup file there must not have been a file to begin with
   elif db_path.isfile():
      db_path.remove()

###

class QuotaManager(object):
   quota_min = 0
   quota_max = 0

   def __init__(self):
      self.quotas = {}

      cmdlib.registerServerCommand('randombots_quota_add', self.addQuota,
       'randombots_quota_add <map name> <min quota> <max quota>\nSpecifies a custom bot quota range for a map.')
      cmdlib.registerServerCommand('randombots_quota_remove', self.removeQuota, 
       'randombots_quota_remove <map name>\nRemoves a custom bota quota range for a map.')

   """ Quota cvar functions """

   def setQuota(self):
      """ Sets a new bot quota """
      # Cancel any existing delay
      self.cancelDelay()
      # If we aren't controlling the bot quota we stop here
      if not cvar_quota_control:
         return

      mapname = str(cvar_currentmap)
      # If the map has custom quotas we use it
      if mapname in self.quotas:
         self.quota_min, self.quota_max = self.quotas[mapname]

      # No custom quotas means we use the default quota range
      else:
         self.quota_min = cvar_quota_default_min
         self.quota_max = cvar_quota_default_max

      # Set random quota
      cvar_quota.set(random.randint(self.quota_min, self.quota_max))

      # Start quota change delay
      gamethread.delayedname(random.randint(cvar_quota_delay_min, cvar_quota_delay_max), 'randombots', self.changeQuota)

   def changeQuota(self):
      """ Changes existing bot quota """
      # Change bot quota
      change = random.choice((-1, 0, 1))
      cvar_quota.set(min(self.quota_max, max(self.quota_min, cvar_quota + change)))
      # Resart quota change delay
      gamethread.delayedname(random.randint(cvar_quota_delay_min, cvar_quota_delay_max), 'randombots', self.changeQuota)

   """ Server command functions """

   def addQuota(self, args):
      """
      randombots_quota_add <map name> <min quota> <max quota>
      """
      if len(args) == 3:
         self.quotas[args[0]] = tuple(int(args[x]) for x in (1, 2))

      else:
         es.dbgmsg(0, 'Syntax: randombots_quota_add <map name> <min quota> <max quota>')

   def removeQuota(self, args):
      """
      randombots_quota_remove <map name>
      """
      if len(args) == 1:
         if args[0] in self.quotas:
            del self.quotas[args[0]]

      else:
         es.dbgmsg(0, 'Syntax: randombots_quota_remove <map name>')

   """ Static functions """

   @staticmethod
   def cancelDelay():
      """ Cancels the quota delay """
      gamethread.cancelDelayed('randombots')

   @staticmethod
   def unregisterCommands():
      """ Unregisters quota server commands """
      cmdlib.unregisterServerCommand('randombots_quota_add')
      cmdlib.unregisterServerCommand('randombots_quota_remove')

quota = QuotaManager()

###

def load():
   config.execute()

   loadProfile()
   selectNextDifficulty()
   quota.setQuota()


def es_map_start(event_var):
   quota.setQuota()
   # Reset the next difficulty for kicks
   selectNextDifficulty()


def player_connect(event_var):
   # When a bot connects we pick a new difficulty for the next bot
   if event_var['networkid'] == 'BOT':
      selectNextDifficulty()


def unload():
   unloadProfile()
   quota.cancelDelay()
   quota.unregisterCommands()

   for cvar in config.getCvars():
      es.ServerVar(cvar).set(0)